## UI Test Tools Framework

A set of UI Test tools to simplify creating WinUI 3 UI Tests. 

### Features:
   - Helpers for waiting for UI to load or update
   - Helpers for grabbing screenshot of UI Elements
   - Helpers for analyzing screenshots using pixel analysis via connected component analysis
   - Code generator to easily generate UI Tests that loads the Window and starts on the correct UI Thread
   
### Installation

1. Create a new `Blank App, Packaged (WinUI 3 in Desktop)` project (not the Windows Application Packaging Project).

  ![image](https://user-images.githubusercontent.com/1378165/145077633-0918e17b-d234-471d-9c08-26e8273a8372.png)

2. Add the `WinUIEx.TestTools.MSTest` NuGet Packaged
3. Add the following line to App.xaml.cs at the end of the "OnLaunched" method: 
```cs
     WinUIEx.TestTools.MSTest.UnitTestClient.Run(m_window);
```
4.  Next add the following package reference to your project file:
```xml
<ItemGroup>
    <PackageReference Include="MSTest.TestAdapter">
        <Version>2.2.8</Version>
    </PackageReference>
</ItemGroup>
```
5. Create a new class to contain your tests, and make the class a `partial` class
6. Import the following namespaces:
```cs
using Microsoft.UI.Xaml;
using Microsoft.UI.Xaml.Controls;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using WinUIEx.TestTools;
using WinUIEx.TestTools.MSTest;
```
7. Mark the class with the `[TestClass]` attribute
8. Add a UI Test method. Example:
```cs
[WinUITestMethod]
public async Task TestGridArrange()
{
    var grid = new Grid() { Background = new Microsoft.UI.Xaml.Media.SolidColorBrush(Microsoft.UI.Colors.Red) };
    WindowContext.Content = grid;
    await grid.LoadAsync();
    Assert.AreEqual(WindowContext.Bounds.Width, grid.ActualWidth, "Full Width");
    Assert.AreEqual(WindowContext.Bounds.Height, grid.ActualHeight, "Full height");
    grid.Width = 300;
    grid.Height = 100;
    await grid.LayoutUpdatedAsync();
    Assert.AreEqual(300, grid.ActualWidth, "ActualWidth");
    Assert.AreEqual(100, grid.ActualHeight, "ActualHeight");

    var bitmap = await grid.AsBitmapAsync();
    Assert.AreEqual(300, bitmap.PixelWidth);
    Assert.AreEqual(100, bitmap.PixelHeight);

    var blobs = await grid.FindConnectedPixelsAsync(Microsoft.UI.Colors.Red);
    Assert.AreEqual(1, blobs.Count);
    Assert.AreEqual(300, blobs[0].Width);
    Assert.AreEqual(100, blobs[0].Height);

    grid.Background = new Microsoft.UI.Xaml.Media.SolidColorBrush(Microsoft.UI.Colors.Green);
    blobs = await grid.FindConnectedPixelsAsync(Microsoft.UI.Colors.Green);
    Assert.AreEqual(1, blobs.Count);
    Assert.AreEqual(300, blobs[0].Width);
    Assert.AreEqual(100, blobs[0].Height);
}
```

### Things to note: 

Class must be `partial` for the UI Test code generator to work

Launching tests with the debugger does not currently work in Visual Studio.
If you need to debug, add the following line of code to launch a debugger after the normal run has started:
```cs
   System.Diagnostics.Debugger.Launch();
```
When this code is hit, a dialog will pop up asking you to pick a VS instance for debugging. Pick the active instance where you have your tests.

New `WindowContext` and `DispatcherQueue` properties are auto-generated by the code generator and gives you access to the current window.
